@@ -40,6 +40,12 @@
 
 #+ccl (setf *print-circle* nil)
 
+
+;;; Without this command encountering unicode characters might cause
+;;; maxima to stop responding on windows.
+#+(and clisp win32) (setf (stream-external-format *socket-connection*) charset:utf-8)
+#+(and clisp win32) (setf custom:*default-file-encoding* charset:utf-8)
+
 ;;; Muffle compiler-notes globally
 #+sbcl (declaim (sb-ext:muffle-conditions sb-ext:compiler-note))
 (defmacro no-warning (form)
@@ -50,7 +56,7 @@
   #+clisp `(let ((custom:*suppress-check-redefinition* t)) ,form)
   #-(or sbcl clisp) `(progn ,form))
 
-($put '$wxmaxima `((mlist simp) 15 8 2) '$version)
+($put '$wxmaxima `((mlist simp) 16 4 1) '$version)
 
 (defun $wxbuild_info ()
   (let ((wxmaxima-version (cdr ($get '$wxmaxima '$version)))
@@ -82,9 +88,6 @@
   ($wxbuild_info)
   (format t "-------------------------------------------------------------~%"))
 
-(setf (get '$inchar 'assign) 'neverset)
-(setf (get '$outchar 'assign) 'neverset)
-
 (defvar *var-tag* '("<v>" "</v>"))
 
 (defun wxxml-get (x p)
@@ -97,6 +100,8 @@
   ;; be inserted
   (setq x (nformat x))
   (cond ((atom x) (wxxml-atom x l r))
+        ((not (listp (car x)))
+         (wxxml (cons '(mlist simp) x) l r lop rop))
         ((or (<= (wxxml-lbp (caar x)) (wxxml-rbp lop))
              (> (wxxml-lbp rop) (wxxml-rbp (caar x))))
          (wxxml-paren x l r))
@@ -135,13 +140,62 @@
   (if (stringp x)
       (let* ((tmp-x (string-substitute "&amp;" #\& x))
 	     (tmp-x (string-substitute "&lt;" #\< tmp-x))
-	     (tmp-x (string-substitute "&gt;" #\> tmp-x)))
+	     (tmp-x (string-substitute "&gt;" #\> tmp-x))
+	     (tmp-x (string-substitute "&#13;" #\Return tmp-x))
+	     (tmp-x (string-substitute "&#13;" #\Linefeed tmp-x))
+	     (tmp-x (string-substitute "&#13;" #\Newline tmp-x)))
 	tmp-x)
       x))
 
 ;;; First we have the functions which are called directly by wxxml and its
 ;;; descendants
 
+(defvar $wxsubscripts nil
+  "Recognize TeX-style subscripts")
+
+(defun $wxdeclare_subscript (x &optional (opt t))
+  (unless (listp x)
+    (setq x (list '(mlist simp) x)))
+  (dolist (s (cdr x))
+    ($put s opt '$wxxml_subscript))
+  opt)
+
+(defun $wxdeclare_subscripted (x &optional (opt t))
+  (unless (listp x)
+    (setq x (list '(mlist simp) x)))
+  (dolist (s (cdr x))
+    ($put s opt '$wxxml_subscripted))
+  opt)
+
+(defun subscriptp (x)
+  (unless (symbolp x)
+    (return-from subscriptp x))
+  (let* ((name (subseq (maybe-invert-string-case (symbol-name x)) 1))
+         (pos (search "_" name :from-end t))
+         #-gcl (*readtable* (copy-readtable nil)))
+    #-gcl (setf (readtable-case *readtable*) :invert)
+    (when pos
+      (let* ((sub (subseq name (+ pos 1)))
+             (sub-var (subseq name 0 pos))
+             (sub-var-symb (read-from-string (concatenate 'string "$" sub-var)))
+             (sub-symb (read-from-string (concatenate 'string "$" sub)))
+             (sub-int (ignore-errors
+                        (parse-integer sub))))
+        (when (and (> (length sub-var) 0)
+                   (or sub-int
+                       (eq $wxsubscripts '$all)
+                       (= (length sub) 1)
+                       (= (length sub-var) 1)
+                       ($get x '$wxxml_subscripted)
+                       ($get sub-symb '$wxxml_subscript)))
+          (format nil  "<i altCopy=\"~{~a~}\"><r>~a</r><r>~a</r></i>"
+                  (mstring x)
+                  (or (get sub-var-symb 'wxxmlword)
+                      (format nil "<v>~a</v>" sub-var))
+                  (if sub-int
+                      (format nil "<n>~a</n>" sub-int)
+                      (format nil "<v>~a</v>" sub))))))))
+
 (defun wxxml-atom (x l r &aux tmp-x)
   (append l
           (list (cond ((numberp x) (wxxmlnumformat x))
@@ -162,9 +216,11 @@
 		      ((member (type-of x) '(GRAPH DIGRAPH))
 		       (format nil "<v>~a</v>" x))
                       ((typep x 'structure-object)
-		       (format nil "<v>Structure [~A]</v>" (type-of x)))
-		      ((hash-table-p x)
+                       (let ((tmp-string (format nil "~s" x)))
+                         (format nil "<st>~a</st>" (wxxml-fix-string tmp-string))))
+                      ((hash-table-p x)
 		       (format nil "<v>HashTable</v>"))
+                      ((and $wxsubscripts (subscriptp x)))
                       (t (wxxml-stripdollar x))))
 	  r))
 
@@ -352,6 +408,9 @@
 (defprop $conjugate wxxml-matchfix wxxml)
 (defprop $conjugate (("<cj>")"</cj>") wxxmlsym)
 
+(defprop %conjugate wxxml-matchfix wxxml)
+(defprop %conjugate (("<cj>")"</cj>") wxxmlsym)
+
 (defprop mbox wxxml-mbox wxxml)
 (defprop mlabox wxxml-mbox wxxml)
 
@@ -886,6 +945,10 @@
 (defprop %derivative 120. wxxml-lbp)
 (defprop %derivative 119. wxxml-rbp)
 
+(defprop $diff wxxml-derivative wxxml)
+(defprop $diff 120. wxxml-lbp)
+(defprop $diff 119. wxxml-rbp)
+
 (defun wxxml-derivative (x l r)
   (if (and $derivabbrev
            (every #'integerp (odds (cddr x) 0))
@@ -1265,11 +1328,13 @@
 		      (format nil "maxout_~d_~d.png" (getpid) *image-counter*)
 		      (format nil "maxout_~d_~d" (getpid) *image-counter*))))
 
+;; The "solid" has to be changed to "dashed" as soon as plot() starts
+;; supporting other line styles than "solid" or "dots".
 (defun $wxplot_preamble ()
   (let ((frmt 
          (cond
            ($wxplot_old_gnuplot "set terminal png picsize ~d ~d; set zeroaxis;")
-           ($wxplot_pngcairo "set terminal pngcairo dashed background \"white\" enhanced font \"arial,10\" fontscale 1.0 size ~d,~d; set zeroaxis;")
+           ($wxplot_pngcairo "set terminal pngcairo solid background \"white\" enhanced font \"arial,10\" fontscale 1.0 size ~d,~d; set zeroaxis;")
            (t "set terminal png size ~d,~d; set zeroaxis;"))))
     (format nil frmt
 	    ($first $wxplot_size)
