@@ -312,6 +312,9 @@ void GroupCell::RemoveOutput()
 
 void GroupCell::AppendOutput(MathCell *cell)
 {
+  wxASSERT_MSG(cell != NULL,_("Bug: Trying to append NULL to a group cell."));
+  if(cell == NULL) return;
+  cell->SetParentList(this);
   if (m_output == NULL) {
     m_output = cell;
 
@@ -443,6 +446,18 @@ void GroupCell::RecalculateSize(CellParser& parser, int fontsize)
   }
 
   m_appendedCells = NULL;
+
+  if(m_previous == NULL)
+  {
+    m_currentPoint.x = MC_GROUP_LEFT_INDENT;
+    m_currentPoint.y = MC_BASE_INDENT + GetMaxCenter();
+  }
+  else
+  {
+    m_currentPoint.x = MC_GROUP_LEFT_INDENT;
+    m_currentPoint.y = dynamic_cast<GroupCell*>(m_previous) -> m_currentPoint.y +
+      dynamic_cast<GroupCell*>(m_previous) -> GetMaxDrop() + GetMaxCenter() + MC_GROUP_SKIP;
+  }
 }
 
 // We assume that appended cells will be in a new line!
@@ -514,7 +529,7 @@ void GroupCell::Draw(CellParser& parser, wxPoint point, int fontsize)
     //
     // Paint background if we have a text cell
     //
-    if (m_groupType == GC_TYPE_TEXT) {
+    if (m_groupType == GC_TYPE_TEXT && !parser.GetPrinter()) {
       wxRect rect = GetRect(false);
       int y = rect.GetY();
 
@@ -657,38 +672,12 @@ wxString GroupCell::ToString()
   return str;
 }
 
-wxString GroupCell::PrepareForTeX(wxString str)
-{
-#if !wxUSE_UNICODE
-  wxString str1(str.wc_str(wxConvLocal), wxConvUTF8);
-#else
-  wxString str1(str);
-#endif
-
-  str1.Replace(wxT("\\"), wxT("\\verb|\\|"));
-  str1.Replace(wxT("_"), wxT("\\_"));
-  str1.Replace(wxT("%"), wxT("\\%"));
-  str1.Replace(wxT("$"), wxT("\\$"));
-  str1.Replace(wxT("{"), wxT("\\{"));
-  str1.Replace(wxT("}"), wxT("\\}"));
-  str1.Replace(wxT("^"), wxT("\\verb|^|"));
-  str1.Replace(wxT(">"), wxT("\\verb|>|"));
-  str1.Replace(wxT("<"), wxT("\\verb|<|"));
-
-#if !wxUSE_UNICODE
-  wxString str2(str1.wc_str(wxConvUTF8), wxConvLocal);
-#else
-  wxString str2(str1);
-#endif
-
-  return str2;
-}
-
 wxString GroupCell::ToTeX()
 {
   return ToTeX(wxEmptyString, wxEmptyString, NULL);
 }
 
+
 wxString GroupCell::ToTeX(wxString imgDir, wxString filename, int *imgCounter)
 {
   wxString str;
@@ -696,203 +685,213 @@ wxString GroupCell::ToTeX(wxString imgDir, wxString filename, int *imgCounter)
   // Now we might want to introduce some markdown:
   MarkDownTeX MarkDownParser;
 
+  switch (m_groupType)
+  {
+  case GC_TYPE_PAGEBREAK:
+    str = wxT("\\pagebreak\n");
+    SuppressLeadingNewlines = true;
+    break;
+
+  case GC_TYPE_IMAGE:
+    if (imgDir != wxEmptyString) {
+      MathCell *copy = m_output->Copy();
+      (*imgCounter)++;
+      wxString image = filename + wxString::Format(wxT("_%d"), *imgCounter);
+      wxString file = imgDir + wxT("/") + image + wxT(".") + dynamic_cast<ImgCell*>(copy)->GetExtension();
+
+      if (!wxDirExists(imgDir))
+        wxMkdir(imgDir);
+
+      if (dynamic_cast<ImgCell*>(copy)->ToImageFile(file).x>=0)
+      {
+        str << wxT("\\begin{figure}[htb]\n")
+            << wxT("  \\begin{center}\n")
+            << wxT("    \\includeimage{")
+            << filename << wxT("_img/") << image << wxT("}\n")
+            << wxT("  \\caption{") << m_input->m_next->ToTeX() << wxT("}\n")
+            << wxT("  \\end{center}\n")
+            << wxT("\\end{figure}\n");
+      }
+    }
+    else
+      str << wxT("\n\\verb|<<GRAPHICS>>|\n");
+    break;
+
+  case GC_TYPE_CODE:
+    str = ToTeXCodeCell(imgDir, filename, imgCounter);
+    break;
+    
+  default:
+    if (GetEditable() != NULL && !m_hide) {
+      str = GetEditable()->ListToTeX();
+      switch (GetEditable()->GetStyle()) {
+      case TS_TITLE:
+        str = wxT("\n\\pagebreak{}\n{\\Huge {\\sc ") + str + wxT("}}\n");
+        str += wxT("\\setcounter{section}{0}\n\\setcounter{subsection}{0}\n");
+        str += wxT("\\setcounter{figure}{0}\n");
+        break;
+      case TS_SECTION:
+        str = wxT("\n\\section{") + str + wxT("}\n");
+        break;
+      case TS_SUBSECTION:
+        str = wxT("\n\\subsection{") + str + wxT("}\n");
+        break;
+      case TS_SUBSUBSECTION:
+        str = wxT("\n\\subsubsection{") + str + wxT("}\n");
+        break;
+      default:
+        if (str.StartsWith(wxT("TeX:"))) {
+          str = GetEditable()->ToString();
+          str = str.Mid(5, str.Length());
+        }
+        else {
+          str = MarkDownParser.MarkDown(str);
+        }
+        break;
+      }
+    }
+    break;
+  }
+  
+  return str;
+}
+
+wxString GroupCell::ToTeXCodeCell(wxString imgDir, wxString filename, int *imgCounter)
+{
+  wxString str;
   bool exportInput = true;
   wxConfig::Get()->Read(wxT("exportInput"), &exportInput);
 
-  // pagebreak
-  if (m_groupType == GC_TYPE_PAGEBREAK) {
-    str = wxT("\\pagebreak\n");
-    SuppressLeadingNewlines = true;
-  }
-
-  // IMAGE CELLS
-  else if (m_groupType == GC_TYPE_IMAGE && imgDir != wxEmptyString) {
-    MathCell *copy = m_output->Copy();
-    (*imgCounter)++;
-    wxString image = filename + wxString::Format(wxT("_%d"), *imgCounter);
-    wxString file = imgDir + wxT("/") + image + wxT(".png");
-
-    Bitmap bmp;
-    bmp.SetData(copy);
-
-    if (!wxDirExists(imgDir))
-      wxMkdir(imgDir);
-
-    if (bmp.ToFile(file).x>=0)
+  // Input cells
+  if(exportInput)
+  {
+    str = wxT("\n\n\\noindent\n%%%%%%%%%%%%%%%\n")
+      wxT("%%% INPUT:\n")
+      wxT("\\begin{minipage}[t]{8ex}\\color{red}\\bf\n") +
+      m_input->ToTeX() +
+      wxT("\n\\end{minipage}");      
+    if (m_input->m_next!=NULL)
     {
-      str << wxT("\\begin{figure}[htb]\n")
-          << wxT("  \\begin{center}\n")
-          << wxT("    \\includeimage{")
-          << filename << wxT("_img/") << image << wxT("}\n")
-          << wxT("  \\caption{") << PrepareForTeX(m_input->m_next->GetValue()) << wxT("}\n")
-          << wxT("  \\end{center}\n")
-          << wxT("\\end{figure}");
-    }
-  }
-  else if (m_groupType == GC_TYPE_IMAGE)
-    str << wxT("\n\\vert|<<GRAPHICS>>|\n");
-
-  // CODE CELLS
-  else if (m_groupType == GC_TYPE_CODE) {
-    // Input cells
-    if(exportInput)
-    {
-      str = wxT("\n\\noindent\n%%%%%%%%%%%%%%%\n")
-        wxT("%%% INPUT:\n")
-        wxT("\\begin{minipage}[t]{8ex}\\color{red}\\bf\n")
-        wxT("\\begin{verbatim}\n") +
-        m_input->ToString() +
-        wxT("\n\\end{verbatim}\n\\end{minipage}");
       
-      if (m_input->m_next!=NULL)
-      {
-        
-        wxString input = m_input->m_next->ToString();
-#if wxUSE_UNICODE
-        input.Replace(wxT("\x2212"), wxT("-")); // unicode minus sign
-        input.Replace(wxT("\xDCB6"), wxT(" ")); // Some weird unicode space character
-#endif
-        str += wxT("\n\\begin{minipage}[t]{\\textwidth}\\color{blue}\n\\begin{verbatim}\n") +
-          input +
-             wxT("\n\\end{verbatim}\n\\end{minipage}");
-      }
-      
-      str += wxT("\n");
-    }
-    else str = wxEmptyString;
-
-    if (m_output != NULL) {
-      str += wxT("%%% OUTPUT:\n");
-      // Need to define labelcolor if this is Copy as LaTeX!
-      if (imgCounter == NULL)
-        str += wxT("\\definecolor{labelcolor}{RGB}{100,0,0}\n");
-      MathCell *tmp = m_output;
-
-      bool mathMode = false;
-
-      while (tmp != NULL)
-      {
-        if (tmp->GetType() == MC_TYPE_IMAGE || tmp->GetType() == MC_TYPE_SLIDE)
-        {
-          if (imgDir != wxEmptyString)
-          {
-            MathCell *copy = tmp->Copy();
-            (*imgCounter)++;
-            wxString image = filename + wxString::Format(wxT("_%d"), *imgCounter);
-	    
-            if (!wxDirExists(imgDir))
-              if (!wxMkdir(imgDir))
-                continue;
-	    
-	    // Do we want to output LaTeX animations?
-	    bool AnimateLaTeX=true;
-	    wxConfig::Get()->Read(wxT("AnimateLaTeX"), &AnimateLaTeX);
-	    if((tmp->GetType() == MC_TYPE_SLIDE)&&(AnimateLaTeX))
-            {
-              SlideShow* src=(SlideShow *)tmp;
-              str << wxT("\\begin{animateinline}{")+wxString::Format(wxT("%i"), src->GetFrameRate())+wxT("}\n");
-              for(int i=0;i<src->Length();i++)
-              {
-                wxString Frame = imgDir + wxT("/") + image + wxString::Format(wxT("_%i"), i);
-                if((src->GetBitmap(i)).SaveFile(Frame+wxT(".png")))
-                  str << wxT("\\includegraphics[width=.95\\linewidth,height=.80\\textheight,keepaspectratio]{")+Frame+wxT("}\n");
-                else
-                  str << wxT("\n\\verb|<<GRAPHICS>>|\n");
-                if(i<src->Length()-1)
-                  str << wxT("\\newframe");
-              }
-              str << wxT("\\end{animateinline}");
-            }
-	    else
-            {
-              wxString file = imgDir + wxT("/") + image + wxT(".png");
-              
-              Bitmap bmp;
-              bmp.SetData(copy);
-              if (bmp.ToFile(file).x>=0)
-                str += wxT("\\includegraphics[width=.95\\linewidth,height=.80\\textheight,keepaspectratio]{") +
-                  filename + wxT("_img/") + image + wxT("}");
-              else
-                str << wxT("\n\\verb|<<GRAPHICS>>|\n");
-            }
-	  }
-	}
-        else if (tmp->GetStyle() == TS_LABEL)
-        {
-	  if(mathMode)
-	    str += wxT("\\]\n\\[\\displaystyle\n");
-	  else
-	    {
-	      str += wxT("\n\\[\\displaystyle\n");
-	      mathMode=true;
-	    }
-          str += wxT("\\parbox{10ex}{$\\color{labelcolor}") + tmp->ToTeX() + wxT("$}\n");
-        }
-	
-        else
-	  {
-	    if((tmp->GetStyle() == TS_DEFAULT)||(tmp->GetStyle() == TS_STRING))
-	      {
-		// We either got a bracket or a parenthesis or regular Text
-		// that shouldn't be displayed as math.
-		if((mathMode)&&(tmp->ToString().Length()>2))
-		  {
-		    str += wxT("\\mbox{}");
-		    str += wxT("\n\\]\n");
-		    mathMode = false;
-		  }
-	      }
-	    else
-	      {
-		if(!mathMode)
-		  {
-		    str += wxT("\n\n\\[\\displaystyle\n");
-		    mathMode = true;
-		  }
-	      }		
-	    str += TexEscapeOutputCell(tmp->ToTeX());
-	  }
-        tmp = tmp->m_nextToDraw;
-      }
-      if(mathMode)
-	{
-	  // Some invisible dummy content that keeps TeX happy if there really is
-	  // no output to display.
-	  str += wxT("\\mbox{}");
-	  str += wxT("\n\\]\n%%%%%%%%%%%%%%%\n");
-	}
-    }
+      wxString input = m_input->m_next->ToTeX();
+      str += wxT("\n\\begin{minipage}[t]{\\textwidth}\\color{blue}\n") +
+        input +
+        wxT("\n\\end{minipage}");
+    }      
   }
 
-  // TITLES, SECTIONS, SUBSECTIONS, SUBSUBSECTIONS, TEXT
-  else if (GetEditable() != NULL && !m_hide) {
-    str = GetEditable()->ListToTeX();
-    switch (GetEditable()->GetStyle()) {
-      case TS_TITLE:
-        str = wxT("\n\\pagebreak{}\n{\\Huge {\\sc ") + PrepareForTeX(str) + wxT("}}\n");
-        str += wxT("\\setcounter{section}{0}\n\\setcounter{subsection}{0}\n");
-        str += wxT("\\setcounter{figure}{0}\n\n");
-        break;
-      case TS_SECTION:
-        str = wxT("\n\\section{") + PrepareForTeX(str) + wxT("}\n\n");
-        break;
-      case TS_SUBSECTION:
-        str = wxT("\n\\subsection{") + PrepareForTeX(str) + wxT("}\n\n");
-        break;
-      case TS_SUBSUBSECTION:
-        str = wxT("\n\\subsubsection{") + PrepareForTeX(str) + wxT("}\n\n");
-        break;
-      default:
-        if (str.StartsWith(wxT("TeX:")))
-          str = str.Mid(5, str.Length());
-        else {
-          str = PrepareForTeX(str);
-	  str = MarkDownParser.MarkDown(str);
-        }
-        break;
-    }
-  }
+  if (m_output != NULL) {
+    str += wxT("\n%%% OUTPUT:\n");
+    // Need to define labelcolor if this is Copy as LaTeX!
+    if (imgCounter == NULL)
+      str += wxT("\\definecolor{labelcolor}{RGB}{100,0,0}\n");
     
+    MathCell *tmp = m_output;
+    
+    bool mathMode = false;
+    
+    while (tmp != NULL)
+    {
+
+      if (tmp->GetType() == MC_TYPE_IMAGE ||
+          tmp->GetType() == MC_TYPE_SLIDE)
+      {
+        str << ToTeXImage(tmp, imgDir, filename, imgCounter);
+      }
+      else
+      {
+        switch (tmp->GetStyle())
+        {
+      
+        case TS_LABEL:
+        case TS_USERLABEL:
+          if(mathMode)
+          str += wxT("\\mbox{}\\]\n\\[\\displaystyle\n");
+          else
+          {
+            str += wxT("\\[\\displaystyle\n");
+            mathMode = true;
+          }
+          str += tmp->ToTeX() + wxT("\n");
+          break;
+        
+        case TS_STRING:
+          if (mathMode)
+          {
+            str += wxT("\\mbox{}\n\\]");
+            mathMode = false;
+          }
+          str += TexEscapeOutputCell(tmp->ToTeX()) + wxT("\n");
+          break;
+        
+        default:
+          if(!mathMode)
+          {
+            str += wxT("\\[\\displaystyle\n");
+            mathMode = true;
+          }
+          str += tmp->ToTeX();
+          break;
+        }
+      }
+      
+      tmp = tmp->m_nextToDraw;
+    }
+    
+    if(mathMode)
+    {
+      // Some invisible dummy content that keeps TeX happy if there really is
+      // no output to display.
+      str += wxT("\\mbox{}\n\\]\n%%%%%%%%%%%%%%%");
+    }
+  }
+  
+  return str;
+}
+
+wxString GroupCell::ToTeXImage(MathCell *tmp, wxString imgDir, wxString filename, int *imgCounter)
+{
+  wxString str;
+  
+  if (imgDir != wxEmptyString)
+  {
+    MathCell *copy = tmp->Copy();
+    (*imgCounter)++;
+    wxString image = filename + wxString::Format(wxT("_%d"), *imgCounter);	    
+    if (!wxDirExists(imgDir))
+      if (!wxMkdir(imgDir))
+        return wxEmptyString;
+    
+    // Do we want to output LaTeX animations?
+    bool AnimateLaTeX=true;
+    wxConfig::Get()->Read(wxT("AnimateLaTeX"), &AnimateLaTeX);
+    if((tmp->GetType() == MC_TYPE_SLIDE)&&(AnimateLaTeX))
+    {
+      SlideShow* src=(SlideShow *)tmp;
+      str << wxT("\\begin{animateinline}{")+wxString::Format(wxT("%i"), src->GetFrameRate())+wxT("}\n");
+      for(int i=0;i<src->Length();i++)
+      {
+        wxString Frame = imgDir + wxT("/") + image + wxString::Format(wxT("_%i"), i);
+        if((src->GetBitmap(i)).SaveFile(Frame+wxT(".png")))
+          str << wxT("\\includegraphics[width=.95\\linewidth,height=.80\\textheight,keepaspectratio]{")+Frame+wxT("}\n");
+        else
+          str << wxT("\n\\verb|<<GRAPHICS>>|\n");
+        if(i<src->Length()-1)
+          str << wxT("\\newframe");
+      }
+      str << wxT("\\end{animateinline}");
+    }
+    else
+    {
+      wxString file = imgDir + wxT("/") + image + wxT(".") + dynamic_cast<ImgCell*>(copy)->GetExtension();
+      if (dynamic_cast<ImgCell*>(copy)->ToImageFile(file).x>=0)
+        str += wxT("\\includegraphics[width=.95\\linewidth,height=.80\\textheight,keepaspectratio]{") +
+          filename + wxT("_img/") + image + wxT("}");
+      else
+        str << wxT("\n\\verb|<<GRAPHICS>>|\n");
+    }
+  }
+
   return str;
 }
 
@@ -1164,7 +1163,7 @@ void GroupCell::SelectOutput(MathCell **start, MathCell **end)
 
   *start = m_output;
 
-  while (*start != NULL && (*start)->GetStyle() != TS_LABEL)
+  while (*start != NULL && ((*start)->GetStyle() != TS_LABEL) && ((*start)->GetStyle() != TS_USERLABEL))
     *start = (*start)->m_nextToDraw;
 
 
@@ -1225,6 +1224,9 @@ void GroupCell::Hide(bool hide) {
   if ((m_groupType == GC_TYPE_TEXT) || (m_groupType == GC_TYPE_CODE))
     GetEditable()->SetFirstLineOnly(m_hide);
 
+  // Don't keep cached versions of scaled images around if they aren't visible at all.
+  GetLabel()->ClearCacheList();
+
   ResetSize();
   GetEditable()->ResetSize();
 }
@@ -1242,6 +1244,15 @@ bool GroupCell::HideTree(GroupCell *tree)
     return false;
   m_hiddenTree = tree;
   m_hiddenTree->SetHiddenTreeParent(this);
+
+  // Clear cached images from cells that are hidden
+  GroupCell *tmp = m_hiddenTree;
+  while(tmp)
+  {
+    tmp->GetLabel()->ClearCacheList();
+    tmp = dynamic_cast<GroupCell *>(tmp->m_next);
+  }
+  
   return true;
 }
 
@@ -1294,6 +1305,8 @@ GroupCell *GroupCell::Fold() {
   GroupCell *start = end; // first to fold
 
   while (end) {
+    end->GetLabel()->ClearCacheList();
+ 
     GroupCell *tmp = dynamic_cast<GroupCell*>(end->m_next);
     if (tmp == NULL)
       break;
@@ -1484,10 +1497,9 @@ bool GroupCell::Contains(GroupCell *cell)
       return true;
 
     // If this cell contains a hidden tree we have to search that at well.
-    if (this->IsFoldable())
-      if(this->GetHiddenTree())
-        if (this->GetHiddenTree()->Contains(cell))
-          return true;
+    if ((tmp->IsFoldable()) && (tmp->GetHiddenTree()) )
+      if (this->GetHiddenTree()->Contains(cell))
+        return true;
 
     // Step to the next cell.
     tmp= dynamic_cast<GroupCell*>(tmp->m_next);
