@@ -1,7 +1,7 @@
 // -*- mode: c++; c-file-style: "linux"; c-basic-offset: 2; indent-tabs-mode: nil -*-
 //
 //  Copyright (C) 2004-2015 Andrej Vodopivec <andrej.vodopivec@gmail.com>
-//            (C) 2004-2015 Gunter Königsmann <wxMaxima@physikbuch.de>
+//            (C) 2014-2016 Gunter Königsmann <wxMaxima@physikbuch.de>
 //
 //  This program is free software; you can redistribute it and/or modify
 //  it under the terms of the GNU General Public License as published by
@@ -48,14 +48,39 @@
 #include "SlideShowCell.h"
 #include "GroupCell.h"
 
+wxXmlNode* MathParser::SkipWhitespaceNode(wxXmlNode* node)
+{
+  if(node)
+  {
+    // If this is a text node there is a chance that this is a whitespace we want to skip
+    if(node->GetType() == wxXML_TEXT_NODE)
+    {
+      // This is a text node => Let's see if it is whitespace-only and skip it if it is.
+      wxString contents = node->GetContent();
+      contents.Trim();
+      if(contents.Length()<=1)
+        node = node->GetNext();
+    }
+  }
+  return node;
+}
+
+wxXmlNode* MathParser::GetNextTag(wxXmlNode* node)
+{
+  if(node)
+    node = node->GetNext();
+  return SkipWhitespaceNode(node);
+}
+
 MathParser::MathParser(wxString zipfile)
 {
+  m_workingDirectory = wxEmptyString;
   m_ParserStyle = MC_TYPE_DEFAULT;
   m_FracStyle = FracCell::FC_NORMAL;
   m_highlight = false;
   if (zipfile.Length() > 0) {
     m_fileSystem = new wxFileSystem();
-    m_fileSystem->ChangePathTo(wxT("file:") + zipfile + wxT("#zip:/"), true);
+    m_fileSystem->ChangePathTo(zipfile + wxT("#zip:/"), true);
   }
   else
     m_fileSystem = NULL;
@@ -86,6 +111,7 @@ MathCell* MathParser::ParseCellTag(wxXmlNode* node)
   if (type == wxT("code")) {
     group = new GroupCell(GC_TYPE_CODE);
     wxXmlNode *children = node->GetChildren();
+    children = SkipWhitespaceNode(children);
     while (children) {
       if (children->GetName() == wxT("input")) {
         MathCell *editor = ParseTag(children->GetChildren());
@@ -95,14 +121,14 @@ MathCell* MathParser::ParseCellTag(wxXmlNode* node)
       if (children->GetName() == wxT("output"))
       {
         MathCell *tag = ParseTag(children->GetChildren());
-        if(tag != NULL)
-          group->AppendOutput(tag);
+        group->AppendOutput(tag);
       }
-      children = children->GetNext();
+      children = GetNextTag(children);
     }
   }  else if (type == wxT("image")) {
     group = new GroupCell(GC_TYPE_IMAGE);
     wxXmlNode *children = node->GetChildren();
+    children = SkipWhitespaceNode(children);
     while (children) {
       if (children->GetName() == wxT("editor")) {
         MathCell *ed = ParseEditorTag(children);
@@ -111,7 +137,7 @@ MathCell* MathParser::ParseCellTag(wxXmlNode* node)
       }
       else
         group->AppendOutput(ParseTag(children));
-      children = children->GetNext();
+      children = GetNextTag(children);
     }
   }
   else if (type == wxT("pagebreak")) {
@@ -149,6 +175,7 @@ MathCell* MathParser::ParseCellTag(wxXmlNode* node)
       return NULL;
 
     wxXmlNode *children = node->GetChildren();
+    children = SkipWhitespaceNode(children);
     while (children) {
       if (children->GetName() == wxT("editor")) {
         MathCell *ed = ParseEditorTag(children);
@@ -157,24 +184,28 @@ MathCell* MathParser::ParseCellTag(wxXmlNode* node)
       }
       else if (children->GetName() == wxT("fold")) { // we have folded groupcells
         wxXmlNode *xmlcells = children->GetChildren();
+        xmlcells = SkipWhitespaceNode(xmlcells);
         MathCell *tree = NULL;
         MathCell *last = NULL;
-        if (xmlcells) {
-          last = tree = ParseTag(xmlcells, false); // first cell
-          while (xmlcells->GetNext()) {
-            xmlcells = xmlcells->GetNext();
-            MathCell *cell = ParseTag(xmlcells, false);
+        while (xmlcells) {
+          MathCell *cell = ParseTag(xmlcells, false);
 
+          if(tree == NULL) tree = cell;
+
+          if(last == NULL) last = cell;
+          else
+          {
             last->m_next = last->m_nextToDraw = cell;
             last->m_next->m_previous = last->m_next->m_previousToDraw = last;
-
+            
             last = last->m_next;
           }
-          if (tree)
-            group->HideTree((GroupCell *)tree);
+          xmlcells = GetNextTag(xmlcells);
         }
+        if (tree)
+          group->HideTree((GroupCell *)tree);
       }
-      children = children->GetNext();
+      children = GetNextTag(children);
     }
   }
 
@@ -226,23 +257,31 @@ MathCell* MathParser::ParseFracTag(wxXmlNode* node)
   frac->SetFracStyle(m_FracStyle);
   frac->SetHighlight(m_highlight);
   wxXmlNode* child = node->GetChildren();
+  child = SkipWhitespaceNode(child);
   if (child)
   {
     frac->SetNum(ParseTag(child, false));
-    child = child->GetNext();
+    child = GetNextTag(child);
+
     if (child)
     {
       frac->SetDenom(ParseTag(child, false));
-      if (node->GetAttributes() != NULL)
+      frac->SetStyle(TS_VARIABLE);
+
+      if(node->GetAttribute(wxT("line")) == wxT("no"))
       {
         frac->SetFracStyle(FracCell::FC_CHOOSE);
       }
+      if(node->GetAttribute(wxT("diffstyle")) == wxT("yes"))
+        frac->SetFracStyle(FracCell::FC_DIFF);
       frac->SetType(m_ParserStyle);
-      frac->SetStyle(TS_VARIABLE);
       frac->SetupBreakUps();
       return frac;
     }
   }
+
+  // The frac tag didn't have a child tag. TODO: Raise an assert in this
+  // case?
   delete frac;
   return NULL;
 }
@@ -251,13 +290,14 @@ MathCell* MathParser::ParseDiffTag(wxXmlNode* node)
 {
   DiffCell *diff = new DiffCell;
   wxXmlNode* child = node->GetChildren();
+  child = SkipWhitespaceNode(child);
   if (child)
   {
     int fc = m_FracStyle;
     m_FracStyle = FracCell::FC_DIFF;
     diff->SetDiff(ParseTag(child, false));
     m_FracStyle = fc;
-    child = child->GetNext();
+    child = GetNextTag(child);
     if (child)
     {
       diff->SetBase(ParseTag(child, true));
@@ -276,10 +316,11 @@ MathCell* MathParser::ParseSupTag(wxXmlNode* node)
   if (node->GetAttributes() != NULL)
     expt->IsMatrix(true);
   wxXmlNode* child = node->GetChildren();
+  child = SkipWhitespaceNode(child);
   if (child)
   {
     expt->SetBase(ParseTag(child, false));
-    child = child->GetNext();
+    child = GetNextTag(child);
     if (child)
     {
       MathCell* power = ParseTag(child, false);
@@ -290,6 +331,9 @@ MathCell* MathParser::ParseSupTag(wxXmlNode* node)
       return expt;
     }
   }
+
+  // SupTag without subscript => Return nothing. TODO: Raise an assert
+  // in this case?
   delete expt;
   return NULL;
 }
@@ -298,16 +342,17 @@ MathCell* MathParser::ParseSubSupTag(wxXmlNode* node)
 {
   SubSupCell *subsup = new SubSupCell;
   wxXmlNode* child = node->GetChildren();
+  child = SkipWhitespaceNode(child);
   if (child)
   {
     subsup->SetBase(ParseTag(child, false));
-    child = child->GetNext();
+    child = GetNextTag(child);
     if (child)
     {
       MathCell* index = ParseTag(child, false);
       index->SetExponentFlag();
       subsup->SetIndex(index);
-      child = child->GetNext();
+      child = GetNextTag(child);
       if (child)
       {
         MathCell* power = ParseTag(child, false);
@@ -319,6 +364,9 @@ MathCell* MathParser::ParseSubSupTag(wxXmlNode* node)
       }
     }
   }
+
+  // subsup tag without super- or subscript => return nothing.
+  // TODO: Raise an assert in this case?
   delete subsup;
   return NULL;
 }
@@ -327,20 +375,30 @@ MathCell* MathParser::ParseSubTag(wxXmlNode* node)
 {
   SubCell *sub = new SubCell;
   wxXmlNode* child = node->GetChildren();
+  child = SkipWhitespaceNode(child);
+  
   if (child)
   {
     sub->SetBase(ParseTag(child, false));
-    child = child->GetNext();
+    child = GetNextTag(child);
     if (child)
     {
       MathCell* index = ParseTag(child, false);
-      index->SetExponentFlag();
-      sub->SetIndex(index);
+      if(index != NULL)
+      {
+        sub->SetIndex(index);
+        index->SetExponentFlag();
+      }
+      else
+        return(NULL);
       sub->SetType(m_ParserStyle);
       sub->SetStyle(TS_VARIABLE);
       return sub;
     }
   }
+
+  // subtag without subscript => Return nothing:
+  // TODO: Raise an assert?
   delete sub;
   return NULL;
 }
@@ -349,11 +407,13 @@ MathCell* MathParser::ParseAtTag(wxXmlNode* node)
 {
   AtCell *at = new AtCell;
   wxXmlNode* child = node->GetChildren();
+  child = SkipWhitespaceNode(child);
+
   if (child)
   {
     at->SetBase(ParseTag(child, false));
     at->SetHighlight(m_highlight);
-    child = child->GetNext();
+    child = GetNextTag(child);
     if (child)
     {
       at->SetIndex(ParseTag(child, false));
@@ -370,10 +430,12 @@ MathCell* MathParser::ParseFunTag(wxXmlNode* node)
 {
   FunCell *fun = new FunCell;
   wxXmlNode* child = node->GetChildren();
+  child = SkipWhitespaceNode(child);
+
   if (child)
   {
     fun->SetName(ParseTag(child, false));
-    child = child->GetNext();
+    child = GetNextTag(child);
     if (child)
     {
       fun->SetType(m_ParserStyle);
@@ -382,15 +444,17 @@ MathCell* MathParser::ParseFunTag(wxXmlNode* node)
       return fun;
     }
   }
+  // Function tag without parameters => Return nothing.
+  // TODO: Raise an assert?
   delete fun;
   return NULL;
 }
 
 MathCell* MathParser::ParseText(wxXmlNode* node, int style)
 {
-  TextCell* cell = new TextCell;
   wxString str;
-  if (node != NULL && (str = node->GetContent()) != wxEmptyString)
+  TextCell *retval = NULL;
+  if ((node != NULL) && ((str = node->GetContent()) != wxEmptyString))
   {
 #if !wxUSE_UNICODE
     wxString str1(str.wc_str(wxConvUTF8), *wxConvCurrent);
@@ -415,22 +479,50 @@ MathCell* MathParser::ParseText(wxXmlNode* node, int style)
 	  //	  str = str.Left(left)+wxT("...");
 	}
     }
-    if(style != TS_ERROR)
-      cell->SetType(m_ParserStyle);
-    else
-      cell->SetType(MC_TYPE_ERROR);
-    cell->SetStyle(style);
-    cell->SetHighlight(m_highlight);
-    cell->SetValue(str);
+    
+    wxStringTokenizer lines(str, wxT('\n'));
+    while(lines.HasMoreTokens())
+    {
+      TextCell* cell = new TextCell;
+      if(style != TS_ERROR)
+        cell->SetType(m_ParserStyle);
+      else
+        cell->SetType(MC_TYPE_ERROR);
+      cell->SetStyle(style);
+      cell->SetHighlight(m_highlight);
+      cell->SetValue(lines.GetNextToken());
+      if(retval == NULL)
+        retval = cell;
+      else
+      {
+        cell->ForceBreakLine(true);
+        retval->AppendCell(cell);
+      };
+    }
   }
-  return cell;
+
+  if (retval == NULL)
+    retval = new TextCell;
+
+  wxString breaklineattrib;
+  if(node != NULL)
+    breaklineattrib = node->GetAttribute(wxT("breakline"), wxT("false"));
+
+  if(breaklineattrib == wxT("true"))
+  {
+    if(retval)
+    {
+      retval->ForceBreakLine(true);
+    }
+  }
+  return retval;
 }
 
 MathCell* MathParser::ParseCharCode(wxXmlNode* node, int style)
 {
   TextCell* cell = new TextCell;
   wxString str;
-  if (node != NULL && (str = node->GetContent()) != wxEmptyString)
+  if ((node != NULL) && ((str = node->GetContent()) != wxEmptyString))
   {
     long code;
     if (str.ToLong(&code))
@@ -450,6 +542,8 @@ MathCell* MathParser::ParseCharCode(wxXmlNode* node, int style)
 MathCell* MathParser::ParseSqrtTag(wxXmlNode* node)
 {
   wxXmlNode* child = node->GetChildren();
+  child = SkipWhitespaceNode(child);
+  
   SqrtCell* cell = new SqrtCell;
   cell->SetInner(ParseTag(child, true));
   cell->SetType(m_ParserStyle);
@@ -461,6 +555,7 @@ MathCell* MathParser::ParseSqrtTag(wxXmlNode* node)
 MathCell* MathParser::ParseAbsTag(wxXmlNode* node)
 {
   wxXmlNode* child = node->GetChildren();
+  child = SkipWhitespaceNode(child);
   AbsCell* cell = new AbsCell;
   cell->SetInner(ParseTag(child, true));
   cell->SetType(m_ParserStyle);
@@ -472,6 +567,7 @@ MathCell* MathParser::ParseAbsTag(wxXmlNode* node)
 MathCell* MathParser::ParseConjugateTag(wxXmlNode* node)
 {
   wxXmlNode* child = node->GetChildren();
+  child = SkipWhitespaceNode(child);
   ConjugateCell* cell = new ConjugateCell;
   cell->SetInner(ParseTag(child, true));
   cell->SetType(m_ParserStyle);
@@ -483,6 +579,7 @@ MathCell* MathParser::ParseConjugateTag(wxXmlNode* node)
 MathCell* MathParser::ParseParenTag(wxXmlNode* node)
 {
   wxXmlNode* child = node->GetChildren();
+  child = SkipWhitespaceNode(child);
   ParenCell* cell = new ParenCell;
   cell->SetInner(ParseTag(child, true), m_ParserStyle);
   cell->SetHighlight(m_highlight);
@@ -496,14 +593,15 @@ MathCell* MathParser::ParseLimitTag(wxXmlNode* node)
 {
   LimitCell *limit = new LimitCell;
   wxXmlNode* child = node->GetChildren();
+  child = SkipWhitespaceNode(child);
   if (child)
   {
     limit->SetName(ParseTag(child, false));
-    child = child->GetNext();
+    child = GetNextTag(child);
     if (child)
     {
       limit->SetUnder(ParseTag(child, false));
-      child = child->GetNext();
+      child = GetNextTag(child);
       if (child)
       {
         limit->SetBase(ParseTag(child, false));
@@ -513,6 +611,9 @@ MathCell* MathParser::ParseLimitTag(wxXmlNode* node)
       }
     }
   }
+
+  // Limit tag with out argument => Return nothing.
+  // TODO: Raise an assert?
   delete limit;
   return NULL;
 }
@@ -521,6 +622,7 @@ MathCell* MathParser::ParseSumTag(wxXmlNode* node)
 {
   SumCell *sum = new SumCell;
   wxXmlNode* child = node->GetChildren();
+  child = SkipWhitespaceNode(child);
   wxString type = node->GetAttribute(wxT("type"), wxT("sum"));
 
   if (type == wxT("prod"))
@@ -529,12 +631,12 @@ MathCell* MathParser::ParseSumTag(wxXmlNode* node)
   if (child)
   {
     sum->SetUnder(ParseTag(child, false));
-    child = child->GetNext();
+    child = GetNextTag(child);
     if (child)
     {
       if (type != wxT("lsum"))
         sum->SetOver(ParseTag(child, false));
-      child = child->GetNext();
+      child = GetNextTag(child);
       if (child)
       {
         sum->SetBase(ParseTag(child, false));
@@ -544,6 +646,8 @@ MathCell* MathParser::ParseSumTag(wxXmlNode* node)
       }
     }
   }
+  // Sum without nothing to sum up => Return nothing.
+  // TODO: Raise an assert?
   delete sum;
   return NULL;
 }
@@ -552,6 +656,7 @@ MathCell* MathParser::ParseIntTag(wxXmlNode* node)
 {
   IntCell *in = new IntCell;
   wxXmlNode* child = node->GetChildren();
+  child = SkipWhitespaceNode(child);
   in->SetHighlight(m_highlight);
   if (node->GetAttributes() == NULL)
   {
@@ -559,15 +664,15 @@ MathCell* MathParser::ParseIntTag(wxXmlNode* node)
     if (child)
     {
       in->SetUnder(ParseTag(child, false));
-      child = child->GetNext();
+      child = GetNextTag(child);
       if (child)
       {
         in->SetOver(ParseTag(child, false));
-        child = child->GetNext();
+        child = GetNextTag(child);
         if (child)
         {
           in->SetBase(ParseTag(child, false));
-          child = child->GetNext();
+          child = GetNextTag(child);
           if (child)
           {
             in->SetVar(ParseTag(child, true));
@@ -584,7 +689,7 @@ MathCell* MathParser::ParseIntTag(wxXmlNode* node)
     if (child)
     {
       in->SetBase(ParseTag(child, false));
-      child = child->GetNext();
+      child = GetNextTag(child);
       if (child)
       {
         in->SetVar(ParseTag(child, true));
@@ -615,16 +720,16 @@ MathCell* MathParser::ParseTableTag(wxXmlNode* node)
   if (node->GetAttribute(wxT("rownames"), wxT("false")) == wxT("true"))
     matrix->RowNames(true);
 
-  wxXmlNode* rows = node->GetChildren();
+  wxXmlNode* rows = SkipWhitespaceNode(node->GetChildren());
   while (rows)
   {
     matrix->NewRow();
-    wxXmlNode* cells = rows->GetChildren();
+    wxXmlNode* cells = SkipWhitespaceNode(rows->GetChildren());
     while (cells)
     {
       matrix->NewColumn();
       matrix->AddNewCell(ParseTag(cells, false));
-      cells = cells->GetNext();
+      cells = GetNextTag(cells);
     }
     rows = rows->GetNext();
   }
@@ -637,263 +742,187 @@ MathCell* MathParser::ParseTableTag(wxXmlNode* node)
 MathCell* MathParser::ParseTag(wxXmlNode* node, bool all)
 {
   //  wxYield();
-  MathCell* tmp = NULL;
+  MathCell* retval = NULL;
   MathCell* cell = NULL;
   bool warning = all;
   wxString altCopy;
 
+  node = SkipWhitespaceNode(node);
+
   while (node)
   {
-    // Parse tags
     if (node->GetType() == wxXML_ELEMENT_NODE)
     {
+      // Parse XML tags. The only other type of element we recognize are text
+      // nodes.
       wxString tagName(node->GetName());
 
+      MathCell *tmp = NULL;
       if (tagName == wxT("v"))
       {               // Variables (atoms)
-        if (cell == NULL)
-          cell = ParseText(node->GetChildren(), TS_VARIABLE);
-        else
-          cell->AppendCell(ParseText(node->GetChildren(), TS_VARIABLE));
+        tmp = ParseText(node->GetChildren(), TS_VARIABLE);
       }
       else if (tagName == wxT("t"))
       {          // Other text
         TextStyle style = TS_DEFAULT;
         if(node->GetAttribute(wxT("type")) == wxT("error"))
           style = TS_ERROR;
-
-        if (cell == NULL)
-          cell = ParseText(node->GetChildren(), style);
-        else
-          cell->AppendCell(ParseText(node->GetChildren(), style));
+        tmp = ParseText(node->GetChildren(), style);
       }
       else if (tagName == wxT("n"))
       {          // Numbers
-        if (cell == NULL)
-          cell = ParseText(node->GetChildren(), TS_NUMBER);
-        else
-          cell->AppendCell(ParseText(node->GetChildren(), TS_NUMBER));
+        tmp = ParseText(node->GetChildren(), TS_NUMBER);
       }
       else if (tagName == wxT("h"))
       {          // Hidden cells (*)
-        MathCell* tmp = ParseText(node->GetChildren());
+        tmp = ParseText(node->GetChildren());
         tmp->m_isHidden = true;
-        if (cell == NULL)
-          cell = tmp;
-        else
-          cell->AppendCell(tmp);
       }
       else if (tagName == wxT("p"))
       {          // Parenthesis
-        if (cell == NULL)
-          cell = ParseParenTag(node);
-        else
-          cell->AppendCell(ParseParenTag(node));
+        tmp = ParseParenTag(node);
       }
       else if (tagName == wxT("f"))
       {               // Fractions
-        if (cell == NULL)
-          cell = ParseFracTag(node);
-        else
-          cell->AppendCell(ParseFracTag(node));
+        tmp = ParseFracTag(node);
       }
       else if (tagName == wxT("e"))
       {          // Exponentials
-        if (cell == NULL)
-          cell = ParseSupTag(node);
-        else
-          cell->AppendCell(ParseSupTag(node));
+        tmp = ParseSupTag(node);
       }
       else if (tagName == wxT("i"))
       {          // Subscripts
-        if (cell == NULL)
-          cell = ParseSubTag(node);
-        else
-          cell->AppendCell(ParseSubTag(node));
+        tmp = ParseSubTag(node);
       }
       else if (tagName == wxT("fn"))
       {         // Functions
-        if (cell == NULL)
-          cell = ParseFunTag(node);
-        else
-          cell->AppendCell(ParseFunTag(node));
+        tmp = ParseFunTag(node);
       }
       else if (tagName == wxT("g"))
       {          // Greek constants
-        MathCell* tmp = ParseText(node->GetChildren(), TS_GREEK_CONSTANT);
-        if (cell == NULL)
-          cell = tmp;
-        else
-          cell->AppendCell(tmp);
+        tmp = ParseText(node->GetChildren(), TS_GREEK_CONSTANT);
       }
       else if (tagName == wxT("s"))
       {          // Special constants %e,...
-        MathCell* tmp = ParseText(node->GetChildren(), TS_SPECIAL_CONSTANT);
-        if (cell == NULL)
-          cell = tmp;
-        else
-          cell->AppendCell(tmp);
+        tmp = ParseText(node->GetChildren(), TS_SPECIAL_CONSTANT);
       }
       else if (tagName == wxT("fnm"))
       {         // Function names
-        MathCell* tmp = ParseText(node->GetChildren(), TS_FUNCTION);
-        if (cell == NULL)
-          cell = tmp;
-        else
-          cell->AppendCell(tmp);
+        tmp = ParseText(node->GetChildren(), TS_FUNCTION);
       }
       else if (tagName == wxT("q"))
       {          // Square roots
-        if (cell == NULL)
-          cell = ParseSqrtTag(node);
-        else
-          cell->AppendCell(ParseSqrtTag(node));
+        tmp = ParseSqrtTag(node);
       }
       else if (tagName == wxT("d"))
       {          // Differentials
-        if (cell == NULL)
-          cell = ParseDiffTag(node);
-        else
-          cell->AppendCell(ParseDiffTag(node));
+        tmp = ParseDiffTag(node);
       }
       else if (tagName == wxT("sm"))
       {         // Sums
-        if (cell == NULL)
-          cell = ParseSumTag(node);
-        else
-          cell->AppendCell(ParseSumTag(node));
+        tmp = ParseSumTag(node);
       }
       else if (tagName == wxT("in"))
       {         // integrals
-        if (cell == NULL)
-          cell = ParseIntTag(node);
-        else
-          cell->AppendCell(ParseIntTag(node));
+        tmp = ParseIntTag(node);
       }
       else if (tagName == wxT("mspace"))
       {
-        if (cell == NULL)
-          cell = new TextCell(wxT(" "));
-        else
-          cell->AppendCell(new TextCell(wxT(" ")));
+        tmp = new TextCell(wxT(" "));
       }
       else if (tagName == wxT("at"))
       {
-        if (cell == NULL)
-          cell = ParseAtTag(node);
-        else
-          cell->AppendCell(ParseAtTag(node));
+        tmp = ParseAtTag(node);
       }
       else if (tagName == wxT("a"))
       {
-        if (cell == NULL)
-          cell = ParseAbsTag(node);
-        else
-          cell->AppendCell(ParseAbsTag(node));
+        tmp = ParseAbsTag(node);
       }
       else if (tagName == wxT("cj"))
       {
-        if (cell == NULL)
-          cell = ParseConjugateTag(node);
-        else
-          cell->AppendCell(ParseConjugateTag(node));
+        tmp = ParseConjugateTag(node);
       }
       else if (tagName == wxT("ie"))
       {
-        if (cell == NULL)
-          cell = ParseSubSupTag(node);
-        else
-          cell->AppendCell(ParseSubSupTag(node));
+        tmp = ParseSubSupTag(node);
       }
       else if (tagName == wxT("lm"))
-      {
-        if (cell == NULL)
-          cell = ParseLimitTag(node);
-        else
-          cell->AppendCell(ParseLimitTag(node));
+      { // A limit tag
+        tmp = ParseLimitTag(node);
       }
       else if (tagName == wxT("r"))
-      {
-        if (cell == NULL)
-          cell = ParseTag(node->GetChildren());
-        else
-          cell->AppendCell(ParseTag(node->GetChildren()));
+      { // A group of tags
+        tmp = ParseTag(node->GetChildren());
       }
       else if (tagName == wxT("tb"))
       {
-        if (cell == NULL)
-          cell = ParseTableTag(node);
-        else
-          cell->AppendCell(ParseTableTag(node));
+        tmp = ParseTableTag(node);
       }
       else if ((tagName == wxT("mth")) || (tagName == wxT("line")))
       {
-        MathCell *tmp = ParseTag(node->GetChildren());
+        tmp = ParseTag(node->GetChildren());
         if (tmp != NULL)
           tmp->ForceBreakLine(true);
         else
           tmp = new TextCell(wxT(" "));
-        if (cell == NULL)
-          cell = tmp;
-        else
-          cell->AppendCell(tmp);
       }
       else if (tagName == wxT("lbl"))
       {
-        MathCell* tmp = ParseText(node->GetChildren(), TS_LABEL);
-        tmp->ForceBreakLine(true);
-        if (cell == NULL)
-          cell = tmp;
+        if (node->GetAttribute(wxT("userdefined"), wxT("no")) != wxT("yes"))
+          tmp = ParseText(node->GetChildren(), TS_LABEL);
         else
-          cell->AppendCell(tmp);
+          tmp = ParseText(node->GetChildren(), TS_USERLABEL);
+        tmp->ForceBreakLine(true);
       }
       else if (tagName == wxT("st"))
       {
-        MathCell* tmp = ParseText(node->GetChildren(), TS_STRING);
-        if (cell == NULL)
-          cell = tmp;
-        else
-          cell->AppendCell(tmp);
+        tmp = ParseText(node->GetChildren(), TS_STRING);
       }
       else if (tagName == wxT("hl"))
       {
         bool highlight = m_highlight;
         m_highlight = true;
-        MathCell* tmp = ParseTag(node->GetChildren());
+        tmp = ParseTag(node->GetChildren());
         m_highlight = highlight;
-        if (cell == NULL)
-          cell = tmp;
-        else
-          cell->AppendCell(tmp);
       }
       else if (tagName == wxT("img"))
       {
+        ImgCell *imageCell;
         wxString filename(node->GetChildren()->GetContent());
 #if !wxUSE_UNICODE
         wxString filename1(filename.wc_str(wxConvUTF8), *wxConvCurrent);
         filename = filename1;
 #endif
 
-        ImgCell *tmp;
-
         if (m_fileSystem) // loading from zip
-          tmp = new ImgCell(filename, false, m_fileSystem);
-        else if (node->GetAttribute(wxT("del"), wxT("yes")) != wxT("no"))
-          tmp = new ImgCell(filename, true, NULL);
+          imageCell = new ImgCell(filename, false, m_fileSystem);
         else
-          tmp = new ImgCell(filename, false, NULL);
+        {
+          if (node->GetAttribute(wxT("del"), wxT("yes")) != wxT("no"))
+            imageCell = new ImgCell(filename, true, NULL);
+          else
+          {
+            // This is the only case show_image() produces ergo this is the only
+            // case we might get a local path
 
+            if(
+              (!wxFileExists(filename)) &&
+              (wxFileExists(m_workingDirectory + wxT("/") + filename))
+              )
+              filename = m_workingDirectory + wxT("/") + filename;
+            
+            imageCell = new ImgCell(filename, false, NULL);
+          }
+        }
+        
         if (node->GetAttribute(wxT("rect"), wxT("true")) == wxT("false"))
-          tmp->DrawRectangle(false);
+          imageCell->DrawRectangle(false);
 
-        if (cell == NULL)
-          cell = tmp;
-        else
-          cell->AppendCell(tmp);
+        tmp = imageCell;
       }
       else if (tagName == wxT("slide"))
       {
-        SlideShow *tmp = new SlideShow(m_fileSystem);
+        SlideShow *slideShow = new SlideShow(m_fileSystem);
         wxString str(node->GetChildren()->GetContent());
         wxArrayString images;
         wxString framerate;
@@ -902,7 +931,7 @@ MathCell* MathParser::ParseTag(wxXmlNode* node, bool all)
         {
           long fr;
           if (framerate.ToLong(&fr))
-            tmp->SetFrameRate(fr);
+            slideShow->SetFrameRate(fr);
         }
         while (tokens.HasMoreTokens()) {
           wxString token = tokens.GetNextToken();
@@ -915,74 +944,80 @@ MathCell* MathParser::ParseTag(wxXmlNode* node, bool all)
             images.Add(token);
           }
         }
-        tmp->LoadImages(images);
+        if(slideShow)
+          slideShow->LoadImages(images);
+        tmp = slideShow;
+      }
+      else if (tagName == wxT("editor"))
+      {
+        tmp = ParseEditorTag(node);
+      }
+      else if (tagName == wxT("cell"))
+      {
+        tmp = ParseCellTag(node);
+      }
+      else if (tagName == wxT("ascii"))
+      {
+        tmp = ParseCharCode(node->GetChildren());
+      }
+      else if (node->GetChildren())
+      {
+        tmp = ParseTag(node->GetChildren());
+      }
+
+      // Append the cell we found (tmp) to the list of cells we parsed so far (cell).
+      if(tmp != NULL)
+      {
+        if (node->GetAttribute(wxT("breakline"), wxT("false")) == wxT("true"))
+          tmp->ForceBreakLine(true);
         if (cell == NULL)
           cell = tmp;
         else
           cell->AppendCell(tmp);
       }
-      else if (tagName == wxT("editor"))
-      {
-        if (cell == NULL)
-          cell = ParseEditorTag(node);
-        else
-          cell->AppendCell(ParseEditorTag(node));
-      }
-      else if (tagName == wxT("cell"))
-      {
-        if (cell == NULL)
-          cell = ParseCellTag(node);
-        else
-          cell->AppendCell(ParseCellTag(node));
-      }
-      else if (tagName == wxT("ascii"))
-      {
-        if (cell == NULL)
-          cell = ParseCharCode(node->GetChildren());
-        else
-          cell->AppendCell(ParseCharCode(node->GetChildren()));
-      }
-      else if (node->GetChildren())
-      {
-        if (cell == NULL)
-          cell = ParseTag(node->GetChildren());
-        else
-          cell->AppendCell(ParseTag(node->GetChildren()));
-      }
     }
-    // Parse text
     else
     {
+      // We didn't get a tag but got a text cell => Parse the text.
       if (cell == NULL)
         cell = ParseText(node);
       else
         cell->AppendCell(ParseText(node));
     }
+    
     if (!all)
       break;
-
+    
     if (cell != NULL)
     {
-      if (tmp == NULL)
-        tmp = cell;
+      // Append the new cell to the return value
+      if (retval == NULL)
+        retval = cell;
       else
         cell = cell->m_next;
     }
     else if (warning)
     {
-      wxMessageBox(_("Parts of the document will not be loaded correctly!"), _("Warning"),
-        wxOK | wxICON_WARNING);
-      warning = false;
+      wxString name;
+      name.Trim(true);
+      name.Trim(false);
+      if(cell != NULL) name = cell->ToString();
+      if(name.Length()!= 0)
+      {
+        wxMessageBox(_("Parts of the document will not be loaded correctly!\nFound unknown XML Tag name "+ name), _("Warning"),
+                     wxOK | wxICON_WARNING);
+        warning = false;
+      }
     }
-
-    if (node->GetAttribute(wxT("altCopy"), &altCopy))
+    
+    if ((cell != NULL) &&(node->GetAttribute(wxT("altCopy"), &altCopy)))
       cell->SetAltCopyText(altCopy);
-
-    node = node->GetNext();
+    
+    node = GetNextTag(node);
   }
-
-  if (tmp != NULL)
-    return tmp;
+  
+  if (retval != NULL)
+    return retval;
   return cell;
 }
 
@@ -1024,7 +1059,7 @@ MathCell* MathParser::ParseLine(wxString s, int style)
   graph.Replace(&s, wxT("?"));
 #endif
 
-  if ((s.Length() < showLength) || (showLength=0))
+  if ((s.Length() < showLength) || (showLength==0))
   {
 
     wxXmlDocument xml;
@@ -1036,7 +1071,7 @@ MathCell* MathParser::ParseLine(wxString s, int style)
     wxStringInputStream xmlStream(su);
 #endif
 
-    xml.Load(xmlStream);
+    xml.Load(xmlStream,wxT("UTF-8"),wxXMLDOC_KEEP_WHITESPACE_NODES);
 
     wxXmlNode *doc = xml.GetRoot();
 
@@ -1045,7 +1080,7 @@ MathCell* MathParser::ParseLine(wxString s, int style)
   }
   else
   {
-    cell = new TextCell(_(" << Expression too long to display! >>"));
+    cell = new TextCell(_(" << Expression longer than allowed by the configuration setting! >>"));
     cell->ForceBreakLine(true);
   }
   return cell;
