@@ -1,7 +1,7 @@
 // -*- mode: c++; c-file-style: "linux"; c-basic-offset: 2; indent-tabs-mode: nil -*-
 //
 //  Copyright (C) 2007-2015 Andrej Vodopivec <andrej.vodopivec@gmail.com>
-//            (C) 2014-2015 Gunter Königsmann <wxMaxima@physikbuch.de>
+//            (C) 2014-2016 Gunter Königsmann <wxMaxima@physikbuch.de>
 //
 //  This program is free software; you can redistribute it and/or modify
 //  it under the terms of the GNU General Public License as published by
@@ -23,6 +23,7 @@
 #define SLIDESHOWCELL_H
 
 #include "MathCell.h"
+#include "Image.h"
 #include <wx/image.h>
 
 #include <wx/filesys.h>
@@ -43,6 +44,12 @@ public:
    */
   SlideShow(wxFileSystem *filesystem = NULL,int framerate = -1);
   ~SlideShow();
+  /*! Remove all cached scaled images from memory
+
+    To be called when the slideshow is outside of the displayed portion 
+    of the screen; The bitmaps will be re-generated when needed.
+   */
+  virtual void ClearCache();
   void Destroy();
   void LoadImages(wxArrayString images);
   MathCell* Copy();
@@ -51,7 +58,7 @@ public:
     *first = *last = this;
   }
   int GetDisplayedIndex() { return m_displayed; }
-  wxImage GetBitmap(int n) { return m_bitmaps[n]->ConvertToImage(); }
+  wxImage GetBitmap(int n) { return m_images[n]->GetUnscaledBitmap().ConvertToImage(); }
   void SetDisplayedIndex(int ind);
   int Length() { return m_size; }
   //! Exports the image the slideshow currently displays
@@ -80,7 +87,7 @@ protected:
   int m_size;
   int m_displayed;
   wxFileSystem *m_fileSystem;
-  vector<wxBitmap*> m_bitmaps;
+  vector<Image*> m_images;
   void RecalculateSize(CellParser& parser, int fontsize);
   void RecalculateWidths(CellParser& parser, int fontsize);
   void Draw(CellParser& parser, wxPoint point, int fontsize);
