@@ -1,7 +1,7 @@
 // -*- mode: c++; c-file-style: "linux"; c-basic-offset: 2; indent-tabs-mode: nil -*-
 //
 //  Copyright (C) 2004-2015 Andrej Vodopivec <andrej.vodopivec@gmail.com>
-//            (C) 2014-2015 Gunter Königsmann <wxMaxima@physikbuch.de>
+//            (C) 2014-2016 Gunter Königsmann <wxMaxima@physikbuch.de>
 //
 //  This program is free software; you can redistribute it and/or modify
 //  it under the terms of the GNU General Public License as published by
@@ -21,6 +21,7 @@
 
 #include "TextCell.h"
 #include "Setup.h"
+#include "wx/config.h"
 
 TextCell::TextCell() : MathCell()
 {
@@ -71,6 +72,20 @@ void TextCell::Destroy()
   m_next = NULL;
 }
 
+wxString TextCell::LabelWidthText()
+{
+  wxString result;
+
+  wxConfig *config = (wxConfig *)wxConfig::Get();
+  int labelWidth = 4;
+  config->Read(wxT("labelWidth"), &labelWidth);
+
+  for(int i=0;i<labelWidth;i++)
+    result += wxT("X");
+
+  return result;
+}
+
 void TextCell::RecalculateWidths(CellParser& parser, int fontsize)
 {
   SetAltText(parser);
@@ -85,15 +100,18 @@ void TextCell::RecalculateWidths(CellParser& parser, int fontsize)
 
     // Labels and prompts are fixed width - adjust font size so that
     // they fit in
-    if ((m_textStyle == TS_LABEL) || (m_textStyle == TS_MAIN_PROMPT)) {
+    if ((m_textStyle == TS_LABEL) || (m_textStyle == TS_USERLABEL) || (m_textStyle == TS_MAIN_PROMPT)) {
 	  // Check for output annotations (/R/ for CRE and /T/ for Taylor expressions)
       if (m_text.Right(2) != wxT("/ "))
-        dc.GetTextExtent(wxT("(\%oXXX)"), &m_width, &m_height);
+        dc.GetTextExtent(wxT("(\%o")+LabelWidthText()+wxT(")"), &m_width, &m_height);
       else
-        dc.GetTextExtent(wxT("(\%oXXX)/R/"), &m_width, &m_height);
+        dc.GetTextExtent(wxT("(\%o")+LabelWidthText()+wxT(")/R/"), &m_width, &m_height);
       m_fontSizeLabel = m_fontSize;
+      wxASSERT_MSG((m_width>0)||(m_text==wxEmptyString),_("The letter \"X\" is of width zero. Installing http://www.math.union.edu/~dpvc/jsmath/download/jsMath-fonts.html and checking \"Use JSmath fonts\" in the configuration dialogue should fix it."));
+      if(m_width < 1) m_width = 10;
       dc.GetTextExtent(m_text, &m_labelWidth, &m_labelHeight);
-      while (m_labelWidth >= m_width) {
+      wxASSERT_MSG((m_labelWidth>0)||(m_text==wxEmptyString),_("Seems like something is broken with the maths font. Installing http://www.math.union.edu/~dpvc/jsmath/download/jsMath-fonts.html and checking \"Use JSmath fonts\" in the configuration dialogue should fix it."));
+      while ((m_labelWidth >= m_width)&&(m_fontSizeLabel > 2)) {
         int fontsize1 = (int) (((double) --m_fontSizeLabel) * scale + 0.5);
         dc.SetFont(wxFont(fontsize1, wxFONTFAMILY_MODERN,
               parser.IsItalic(m_textStyle),
@@ -160,11 +178,11 @@ void TextCell::Draw(CellParser& parser, wxPoint point, int fontsize)
     SetForeground(parser);
 
     /// Labels and prompts have special fontsize
-    if ((m_textStyle == TS_LABEL) || (m_textStyle == TS_MAIN_PROMPT))
+    if ((m_textStyle == TS_LABEL) || (m_textStyle == TS_USERLABEL) || (m_textStyle == TS_MAIN_PROMPT))
     {
       SetFont(parser, m_fontSizeLabel);
       dc.DrawText(m_text,
-                  point.x + SCALE_PX(MC_TEXT_PADDING, scale) + (m_width - m_labelWidth),
+                  point.x + SCALE_PX(MC_TEXT_PADDING, scale),
                   point.y - m_realCenter + (m_height - m_labelHeight)/2);
     }
 
@@ -186,6 +204,12 @@ void TextCell::Draw(CellParser& parser, wxPoint point, int fontsize)
                   point.x + SCALE_PX(MC_TEXT_PADDING, scale),
                   point.y - m_realCenter + SCALE_PX(MC_TEXT_PADDING, scale));
 
+#if wxUSE_UNICODE
+    else if (m_text == wxT("#"))
+      dc.DrawText(wxT("\x2260"),
+                  point.x + SCALE_PX(MC_TEXT_PADDING, scale),
+                  point.y - m_realCenter + SCALE_PX(MC_TEXT_PADDING, scale));
+#endif
     /// This is the default.
     else
     {
@@ -232,44 +256,62 @@ void TextCell::SetFont(CellParser& parser, int fontsize)
   // Use jsMath
   if (m_altJs && parser.CheckTeXFonts())
   {
-    dc.SetFont(wxFont(fontsize1, wxFONTFAMILY_MODERN,
+    wxFont font(fontsize1, wxFONTFAMILY_MODERN,
                       wxFONTSTYLE_NORMAL,
                       parser.IsBold(m_textStyle),
                       parser.IsUnderlined(m_textStyle),
-                      m_texFontname));
+                         m_texFontname);
+    wxASSERT_MSG(font.IsOk(),_("Seems like something is broken with a font. Installing http://www.math.union.edu/~dpvc/jsmath/download/jsMath-fonts.html and checking \"Use JSmath fonts\" in the configuration dialogue should fix it."));
+    dc.SetFont(font);
   }
 
   // We have an alternative symbol
   else if (m_alt)
-    dc.SetFont(wxFont(fontsize1, wxFONTFAMILY_MODERN,
+  {
+    wxFont font(fontsize1, wxFONTFAMILY_MODERN,
                       wxFONTSTYLE_NORMAL,
                       parser.IsBold(m_textStyle),
                       false,
                       m_fontname != wxEmptyString ?
                           m_fontname : parser.GetFontName(m_textStyle),
-                      parser.GetFontEncoding()));
-
+                  parser.GetFontEncoding());
+    wxASSERT_MSG(font.IsOk(),_("Seems like something is broken with a font. Installing http://www.math.union.edu/~dpvc/jsmath/download/jsMath-fonts.html and checking \"Use JSmath fonts\" in the configuration dialogue should fix it."));
+    dc.SetFont(font);
+  }
   // Titles, sections, subsections... - don't underline
   else if ((m_textStyle == TS_TITLE) ||
            (m_textStyle == TS_SECTION) ||
            (m_textStyle == TS_SUBSECTION) ||
            (m_textStyle == TS_SUBSUBSECTION)
     )
-    dc.SetFont(wxFont(fontsize1, wxFONTFAMILY_MODERN,
-                      parser.IsItalic(m_textStyle),
-                      parser.IsBold(m_textStyle),
-                      false,
-                      parser.GetFontName(m_textStyle),
-                      parser.GetFontEncoding()));
-
+  {
+    wxFont font(fontsize1, wxFONTFAMILY_MODERN,
+                parser.IsItalic(m_textStyle),
+                parser.IsBold(m_textStyle),
+                false,
+                parser.GetFontName(m_textStyle),
+                parser.GetFontEncoding());
+    wxASSERT_MSG(font.IsOk(),_("Seems like something is broken with a font. Installing http://www.math.union.edu/~dpvc/jsmath/download/jsMath-fonts.html and checking \"Use JSmath fonts\" in the configuration dialogue should fix it."));
+    dc.SetFont(font);
+  }
   // Default
   else
-    dc.SetFont(wxFont(fontsize1, wxFONTFAMILY_MODERN,
-                      parser.IsItalic(m_textStyle),
-                      parser.IsBold(m_textStyle),
-                      parser.IsUnderlined(m_textStyle),
-                      parser.GetFontName(m_textStyle),
-                      parser.GetFontEncoding()));
+  {
+    wxFont font(fontsize1, wxFONTFAMILY_MODERN,
+                parser.IsItalic(m_textStyle),
+                parser.IsBold(m_textStyle),
+                parser.IsUnderlined(m_textStyle),
+                parser.GetFontName(m_textStyle),
+                parser.GetFontEncoding());
+    wxASSERT_MSG(font.IsOk(),_("Seems like something is broken with a font. Installing http://www.math.union.edu/~dpvc/jsmath/download/jsMath-fonts.html and checking \"Use JSmath fonts\" in the configuration dialogue should fix it."));
+    dc.SetFont(font);
+  }
+  
+  // A fallback if the font we selected is no more installed or isn't working at all.
+  if(!dc.GetFont().IsOk())
+  {
+    dc.SetFont(wxFontInfo(10));
+  }
 }
 
 bool TextCell::IsOperator()
@@ -324,14 +366,152 @@ wxString TextCell::ToString()
     text = wxT("\"") + text + wxT("\"");
     break;
   }
-
+  
   return text;
 }
 
 wxString TextCell::ToTeX()
 {
-  // m_IsHidden is set only for multiplication signs
+  wxString text = m_text;
+  text.Replace(wxT("\\"), wxT("\\ensuremath{\\backslash}"));
+  text.Replace(wxT("<"), wxT("\\ensuremath{<}"));
+  text.Replace(wxT(">"), wxT("\\ensuremath{>}"));
+#if wxUSE_UNICODE
+  text.Replace(wxT("\x2212"), wxT("-")); // unicode minus sign
+  text.Replace(L"\x00B1",wxT("\\ensuremath{\\pm}"));
+  text.Replace(L"\x03B1",wxT("\\ensuremath{\\alpha}"));
+  text.Replace(L"\x00B2",wxT("\\ensuremath{^2}"));
+  text.Replace(L"\x00B3",wxT("\\ensuremath{^3}"));
+  text.Replace(L"\x221A",wxT("\\ensuremath{\\sqrt{}}"));
+  text.Replace(L"\x2148",wxT("\\ensuremath{\\mathbbm{i}}"));
+  text.Replace(L"\x2147",wxT("\\ensuremath{\\mathbbm{e}}"));
+  text.Replace(L"\x210f",wxT("\\ensuremath{\\hbar}"));
+  text.Replace(L"\x2203",wxT("\\ensuremath{\\exists}"));
+  text.Replace(L"\x2204",wxT("\\ensuremath{\\nexists}"));
+  text.Replace(L"\x2208",wxT("\\ensuremath{\\in}"));
+  text.Replace(L"\x21D2",wxT("\\ensuremath{\\Longrightarrow}"));
+  text.Replace(L"\x221e",wxT("\\ensuremath{\\infty}"));
+  text.Replace(L"\x22C0",wxT("\\ensuremath{\\wedge}"));
+  text.Replace(L"\x22C1",wxT("\\ensuremath{\\vee}"));
+  text.Replace(L"\x22bb",wxT("\\ensuremath{\\oplus}"));
+  text.Replace(L"\x22BC",wxT("\\ensuremath{\\overline{\\wedge}}"));
+  text.Replace(L"\x22BB",wxT("\\ensuremath{\\overline{\\vee}}"));
+  text.Replace(L"\x00AC",wxT("\\ensuremath{\\setminus}"));
+  text.Replace(L"\x22C3",wxT("\\ensuremath{\\cup}"));
+  text.Replace(L"\x22C2",wxT("\\ensuremath{\\cap}"));
+  text.Replace(L"\x2286",wxT("\\ensuremath{\\subseteq}"));
+  text.Replace(L"\x2282",wxT("\\ensuremath{\\subset}"));
+  text.Replace(L"\x2288",wxT("\\ensuremath{\\not\\subseteq}"));
+  text.Replace(L"\x0127",wxT("\\ensuremath{\\hbar}"));
+  text.Replace(L"\x0126",wxT("\\ensuremath{\\Hbar}"));
+  text.Replace(L"\x2205",wxT("\\ensuremath{\\emptyset}"));
+  text.Replace(L"\x00BD",wxT("\\ensuremath{\\frac{1}{2}}"));
+  text.Replace(L"\x03B2",wxT("\\ensuremath{\\beta}"));
+  text.Replace(L"\x03B3",wxT("\\ensuremath{\\gamma}"));
+  text.Replace(L"\x03B4",wxT("\\ensuremath{\\delta}"));
+  text.Replace(L"\x03B5",wxT("\\ensuremath{\\epsilon}"));
+  text.Replace(L"\x03B6",wxT("\\ensuremath{\\zeta}"));
+  text.Replace(L"\x03B7",wxT("\\ensuremath{\\eta}"));
+  text.Replace(L"\x03B8",wxT("\\ensuremath{\\theta}"));
+  text.Replace(L"\x03B9",wxT("\\ensuremath{\\iota}"));
+  text.Replace(L"\x03BA",wxT("\\ensuremath{\\kappa}"));
+  text.Replace(L"\x03BB",wxT("\\ensuremath{\\lambda}"));
+  text.Replace(L"\x03BC",wxT("\\ensuremath{\\mu}"));
+  text.Replace(L"\x03BD",wxT("\\ensuremath{\\nu}"));
+  text.Replace(L"\x03BE",wxT("\\ensuremath{\\xi}"));
+  text.Replace(L"\x03BF",wxT("\\ensuremath{\\omnicron}"));
+  text.Replace(L"\x03C0",wxT("\\ensuremath{\\pi}"));
+  text.Replace(L"\x03C1",wxT("\\ensuremath{\\rho}"));
+  text.Replace(L"\x03C3",wxT("\\ensuremath{\\sigma}"));
+  text.Replace(L"\x03C4",wxT("\\ensuremath{\\tau}"));
+  text.Replace(L"\x03C5",wxT("\\ensuremath{\\upsilon}"));
+  text.Replace(L"\x03C6",wxT("\\ensuremath{\\phi}"));
+  text.Replace(L"\x03C7",wxT("\\ensuremath{\\chi}"));
+  text.Replace(L"\x03C8",wxT("\\ensuremath{\\psi}"));
+  text.Replace(L"\x03C9",wxT("\\ensuremath{\\omega}"));
+  text.Replace(L"\x0391",wxT("\\ensuremath{\\Alpha}"));
+  text.Replace(L"\x0392",wxT("\\ensuremath{\\Beta}"));
+  text.Replace(L"\x0393",wxT("\\ensuremath{\\Gamma}"));
+  text.Replace(L"\x0394",wxT("\\ensuremath{\\Delta}"));
+  text.Replace(L"\x0395",wxT("\\ensuremath{\\Epsilon}"));
+  text.Replace(L"\x0396",wxT("\\ensuremath{\\Zeta}"));
+  text.Replace(L"\x0397",wxT("\\ensuremath{\\Eta}"));
+  text.Replace(L"\x0398",wxT("\\ensuremath{\\Theta}"));
+  text.Replace(L"\x0399",wxT("\\ensuremath{\\Iota}"));
+  text.Replace(L"\x039A",wxT("\\ensuremath{\\Kappa}"));
+  text.Replace(L"\x039B",wxT("\\ensuremath{\\Lambda}"));
+  text.Replace(L"\x039C",wxT("\\ensuremath{\\Mu}"));
+  text.Replace(L"\x039D",wxT("\\ensuremath{\\Nu}"));
+  text.Replace(L"\x039E",wxT("\\ensuremath{\\Xi}"));
+  text.Replace(L"\x039F",wxT("\\ensuremath{\\Omnicron}"));
+  text.Replace(L"\x03A0",wxT("\\ensuremath{\\Pi}"));
+  text.Replace(L"\x03A1",wxT("\\ensuremath{\\Rho}"));
+  text.Replace(L"\x03A3",wxT("\\ensuremath{\\Sigma}"));
+  text.Replace(L"\x03A4",wxT("\\ensuremath{\\Tau}"));
+  text.Replace(L"\x03A5",wxT("\\ensuremath{\\Upsilon}"));
+  text.Replace(L"\x03A6",wxT("\\ensuremath{\\Phi}"));
+  text.Replace(L"\x03A7",wxT("\\ensuremath{\\Chi}"));
+  text.Replace(L"\x03A8",wxT("\\ensuremath{\\Psi}"));
+  text.Replace(L"\x03A9",wxT("\\ensuremath{\\Omega}"));
+  text.Replace(L"\x2202",wxT("\\ensuremath{\\partial}"));
+  text.Replace(L"\x222b",wxT("\\ensuremath{\\int}"));
+  text.Replace(L"\x2245",wxT("\\ensuremath{\\approx}"));
+  text.Replace(L"\x221d",wxT("\\ensuremath{\\propto}"));
+  text.Replace(L"\x2260",wxT("\\ensuremath{\\neq}"));
+  text.Replace(L"\x2264",wxT("\\ensuremath{\\leq}"));
+  text.Replace(L"\x2265",wxT("\\ensuremath{\\geq}"));
+  text.Replace(L"\x226A",wxT("\\ensuremath{\\ll}"));
+  text.Replace(L"\x226B",wxT("\\ensuremath{\\gg}"));
+  text.Replace(L"\x220e",wxT("\\ensuremath{\\blacksquare}"));
+  text.Replace(L"\x2263",wxT("\\ensuremath{\\equiv}"));
+  text.Replace(L"\x2211",wxT("\\ensuremath{\\sum}"));
+  text.Replace(L"\x220F",wxT("\\ensuremath{\\prod}"));
+  text.Replace(L"\x2225",wxT("\\ensuremath{\\parallel}"));
+  text.Replace(L"\x27C2",wxT("\\ensuremath{\\bot}"));
+  text.Replace(wxT("~"),wxT("\\ensuremath{\\sim }"));
+  text.Replace(wxT("_"), wxT("\\_ "));
+  text.Replace(wxT("$"), wxT("\\$ "));
+  text.Replace(wxT("%"), wxT("\\%{}"));
+  text.Replace(wxT("&"), wxT("\\& "));
+  text.Replace(wxT("@"), wxT("\\ensuremath{@}"));
+  text.Replace(wxT("#"), wxT("\\ensuremath{\\neq}"));
+  text.Replace(wxT("\xDCB6"), wxT("~")); // A non-breakable space
+  text.Replace(wxT("<"), wxT("\\ensuremath{<}"));
+  text.Replace(wxT(">"), wxT("\\ensuremath{>}"));
+  text.Replace(wxT("\x219D"), wxT("\\ensuremath{\\leadsto}"));
+  text.Replace(wxT("\x2192"), wxT("\\ensuremath{\\rightarrow}"));
+#endif
+
+
+  // m_IsHidden is set for multiplication signs and parenthesis that
+  // don't need to be shown
   if (m_isHidden)
+  {
+    // Normally in TeX the spacing between variable names following each other directly
+    // is chosen to show that this is a multiplication.
+    // But any use of \mathit{} will change to ordinary text spacing which means we need
+    // to add a \, to show that we want to multiply the two long variable names.
+    if((text == wxT("*"))||(text == wxT("\xB7")))
+    {
+      // We have a hidden multiplication sign
+      if(
+        // This multiplication sign is between 2 cells
+        ((m_previous != NULL)&&(m_next != NULL)) &&
+        // These cells are two variable names
+        ((m_previous->GetStyle() == TS_VARIABLE) && (m_next->GetStyle() == TS_VARIABLE)) &&
+        // The variable name prior to this cell has no subscript
+        (!(m_previous->ToString().Contains(wxT('_')))) &&
+        // we will be using \mathit{} for the TeX outout.
+        ((m_next->ToString().Length() > 1) || (m_next->ToString().Length() > 1))
+        )
+        text=wxT("\\,");
+      else
+      text = wxEmptyString;        
+    }
+    else
+      text = wxEmptyString;
+  }
+  else
   {
     /*
       Normally we want to draw a centered dot in this case. But if we
@@ -342,111 +522,194 @@ wxString TextCell::ToTeX()
       parenthesis does contain a product.
     */
     
-    if (m_SuppressMultiplicationDot) return wxEmptyString;
-    
-    // If we want to know if the last element was a "d" we first have to
-    // look if there actually is a last element.
-    if(m_previous)
+    if (m_SuppressMultiplicationDot)
     {
-      if (m_previous->GetStyle() == TS_SPECIAL_CONSTANT && m_previous->ToTeX()==wxT("d"))
-        return wxT("\\,");
-      else
-        return wxT("\\cdot ");
+      text.Replace(wxT("*"),wxT("\\,"));
+      text.Replace(wxT("\xB7"),wxT("\\,"));
     }
     else
-      return wxT("\\cdot ");
+    {
+      // If we want to know if the last element was a "d" we first have to
+      // look if there actually is a last element.
+      if(m_previous)
+      {
+        if (m_previous->GetStyle() == TS_SPECIAL_CONSTANT && m_previous->ToTeX()==wxT("d"))
+        {
+          text.Replace(wxT("*"),wxT("\\,"));
+          text.Replace(wxT("\xB7"),wxT("\\,"));
+        }
+        else
+        {
+          text.Replace(wxT("*"),wxT("\\cdot "));
+          text.Replace(wxT("\xB7"),wxT("\\cdot "));
+        }
+      }
+    }
   }
   
-  if (m_textStyle == TS_GREEK_CONSTANT)
+  if (GetStyle() == TS_GREEK_CONSTANT)
   {
-    if (m_text[0] != '%')
-      return wxT("\\") + m_text;
+    if (text[0] != '%')
+      return wxT("\\") + text;
     else
-      return m_text;
+      return text;
   }
   
-  if (m_textStyle == TS_SPECIAL_CONSTANT)
+  if (GetStyle() == TS_SPECIAL_CONSTANT)
   {
-    if (m_text == wxT("inf"))
+    if (text == wxT("inf"))
       return wxT("\\infty ");
-    else if (m_text == wxT("%e"))
+    else if (text == wxT("%e"))
       return wxT("e");
-    else if (m_text == wxT("%i"))
+    else if (text == wxT("%i"))
       return wxT("i");
-    else if (m_text == wxT("%pi"))
-      return wxT("\\pi ");
+    else if (text == wxT("\\%{}pi"))
+      return wxT("\\ensuremath{\\pi} ");
     else
-      return m_text;
+      return text;
   }
-
-  wxString text = m_text;
   
-  if (m_textStyle == TS_LABEL)
-  {    
-    text = wxT("\\mathrm{\\tt ") + m_text + wxT("}\\quad ");
+  if ((GetStyle() == TS_LABEL)||(GetStyle() == TS_USERLABEL))
+  {
+    wxString conditionalLinebreak;
+    if(m_previous) conditionalLinebreak = wxT("\\]\n\\[");
+    text.Trim(true);
+    wxString label = text.SubString(1,text.Length()-2);
+    text = conditionalLinebreak + wxT("\\tag{") + label + wxT("}");
+    label.Replace(wxT("\\%{}"),wxT(""));
+    text += wxT("\\label{") + label + wxT("}");
   }
   else
   {
-    if (m_textStyle == TS_FUNCTION)
+    if (GetStyle() == TS_FUNCTION)
     {
-      if(m_text!=wxEmptyString)
-        text = wxT("\\mathrm{") + m_text + wxT("}");
+      if(text!=wxEmptyString)
+        text = wxT("\\operatorname{") + text + wxT("}");
     }
-    else if (m_textStyle == TS_VARIABLE)
+    else if (GetStyle() == TS_VARIABLE)
     {
-      if(m_text.Length() > 1)
-        text = wxT("\\mathit{") + m_text + wxT("}");
+      if((text.Length() > 1)&&(text[1]!=wxT('_')))
+        text = wxT("\\mathit{") + text + wxT("}");
+      if (text == wxT("\\%{}pi"))
+        text = wxT("\\ensuremath{\\pi} ");
     }
-    else if (m_textStyle == TS_ERROR)
+    else if (GetStyle() == TS_ERROR)
     {
-      if(m_text.Length() > 1)
-        text = wxT("\\mbox{") + m_text + wxT("}");
+      if(text.Length() > 1)
+        text = wxT("\\mbox{") + text + wxT("}");
     }
-    else if (m_textStyle == TS_DEFAULT)
+    else if (GetStyle() == TS_DEFAULT)
     {
-      if(m_text.Length() > 2)
-        text = wxT("\\mbox{") + m_text + wxT("}");
+      if(text.Length() > 2)
+        text = wxT("\\mbox{") + text + wxT("}");
     }
   }
   
   if (
-    (m_textStyle != TS_FUNCTION) &&
-    (m_textStyle != TS_OUTDATED) &&
-    (m_textStyle != TS_VARIABLE) &&
-    (m_textStyle != TS_NUMBER)   &&
-    (m_textStyle != TS_GREEK_CONSTANT)   &&
-    (m_textStyle != TS_SPECIAL_CONSTANT)
+    (GetStyle() != TS_FUNCTION) &&
+    (GetStyle() != TS_OUTDATED) &&
+    (GetStyle() != TS_VARIABLE) &&
+    (GetStyle() != TS_NUMBER)   &&
+    (GetStyle() != TS_GREEK_CONSTANT)   &&
+    (GetStyle() != TS_SPECIAL_CONSTANT)
     )
     text.Replace(wxT("^"), wxT("\\textasciicircum"));
   
-  if((m_textStyle == TS_DEFAULT)||(m_textStyle == TS_STRING))
+  if((GetStyle() == TS_DEFAULT)||(GetStyle() == TS_STRING))
   {
-    if(m_text.Length()>1)
+    if(text.Length()>1)
     {
       if(((m_forceBreakLine)||(m_breakLine)))
-        //text=wxT("\\ifhmode\\\\\\fi\n")+text;
+        //text=wxT("\\ifhmode\\\\fi\n")+text;
         text = wxT("\\mbox{}\\\\")+text;
-      if(m_textStyle != TS_DEFAULT)
+      if(GetStyle() != TS_DEFAULT)
         text.Replace(wxT(" "), wxT("\\,"));
     }
   }
-  text.Replace(wxT("_"), wxT("\\_"));
-  text.Replace(wxT("$"), wxT("\\$"));
-  text.Replace(wxT("%"), wxT("\\%"));
-  text.Replace(wxT("#"), wxT("\\#"));
-#if wxUSE_UNICODE
-  text.Replace(wxT("\x2212"), wxT("-")); // unicode minus sign
-#endif
+
+  if ((GetStyle() == TS_LABEL)||(GetStyle() == TS_LABEL))
+    text = text + wxT(" ");
   return text;
 }
 
+wxString TextCell::ToMathML()
+{
+  wxString text=m_text;
+  text.Replace(wxT("&"),wxT("&amp;"));
+  text.Replace(wxT("<"),wxT("&lt;"));
+  text.Replace(wxT(">"),wxT("&gt;"));
+  text.Replace(wxT("'"),  wxT("&apos;"));
+  text.Replace(wxT("\""), wxT("&quot;"));
+
+  // If we didn't display a multiplication dot we want to do the same in MathML.
+  if(m_isHidden)
+  {
+    text.Replace(wxT("*"),     wxT("&#8290;"));
+    text.Replace(wxT("\xB7"),  wxT("&#8290;"));
+    if(text != wxT ("&#8290;"))
+      text = wxEmptyString;
+  }
+  text.Replace(wxT("*"),  wxT("\xB7"));
+      
+  switch(GetStyle())
+    {
+    case TS_GREEK_CONSTANT:
+      text = GetGreekStringUnicode();
+    case TS_SPECIAL_CONSTANT:
+    {
+      // The "d" from d/dt can be written as a special unicode symbol. But firefox doesn't
+      // support this currently => Commenting it out.
+      // if((GetStyle() == TS_SPECIAL_CONSTANT) && (text == wxT("d")))
+      //   text = wxT("&#2146;");
+      bool keepPercent = true;
+      wxConfig::Get()->Read(wxT("keepPercent"), &keepPercent);
+      if (!keepPercent) {
+        if (text == wxT("%e"))
+          text = wxT("e");
+        else if (text == wxT("%i"))
+          text = wxT("i");
+      }
+    }
+    case TS_VARIABLE:
+    {
+      bool keepPercent = true;
+      wxConfig::Get()->Read(wxT("keepPercent"), &keepPercent);
+      
+      if (!keepPercent) {
+        if (text == wxT("%pi"))
+          text = wxT("\x03C0");
+      }
+    }
+    case TS_FUNCTION:
+      if(text == wxT("inf"))
+        text = wxT("\x221e");
+      return wxT("<mi>")+text+wxT("</mi>\n");
+      break;
+    case TS_NUMBER:
+      return wxT("<mn>")+text+wxT("</mn>\n");
+      break;
+
+    case TS_LABEL:
+    case TS_USERLABEL:
+      return wxT("<mtext>")+text+wxT("</mtext></mtd><mtd>\n");
+      break;
+
+    case TS_STRING:
+    default:
+      if (text.StartsWith(wxT("\"")))
+        return wxT("<ms>")+text+wxT("</ms>\n");
+      else
+        return wxT("<mo>")+text+wxT("</mo>\n");
+    }
+}
+
 wxString TextCell::ToXML()
 {
   wxString tag;
   wxString flags;
   if(m_isHidden)tag=_T("h");
     else
-      switch(m_textStyle)
+      switch(GetStyle())
 	{
 	case TS_GREEK_CONSTANT: tag=_T("g");break;
 	case TS_SPECIAL_CONSTANT: tag=_T("s");break;
@@ -455,10 +718,16 @@ wxString TextCell::ToXML()
 	case TS_NUMBER: tag=_T("n");break;
 	case TS_STRING: tag=_T("st");break;
 	case TS_LABEL: tag=_T("lbl");break;
+	case TS_USERLABEL: tag=_T("lbl");
+          flags += wxT(" userdefined=\"yes\"");
+          break;
 	default: tag=_T("t");
 	}
 
-  if(m_textStyle == TS_ERROR)
+  if((m_forceBreakLine)&&(GetStyle()!=TS_LABEL)&&(GetStyle()!=TS_USERLABEL))
+    flags += wxT(" breakline=\"true\"");
+    
+  if(GetStyle() == TS_ERROR)
     flags += wxT(" type=\"error\"");
     
   wxString xmlstring = m_text;
@@ -489,11 +758,11 @@ bool TextCell::IsShortNum()
 void TextCell::SetAltText(CellParser& parser)
 {
   m_altJs = m_alt = false;
-  if (m_textStyle == TS_DEFAULT)
+  if (GetStyle() == TS_DEFAULT)
     return ;
 
   /// Greek characters are defined in jsMath, Windows and Unicode
-  if (m_textStyle == TS_GREEK_CONSTANT)
+  if (GetStyle() == TS_GREEK_CONSTANT)
   {
     m_altJs = true;
     m_altJsText = GetGreekStringTeX();
@@ -686,7 +955,7 @@ wxString TextCell::GetSymbolUnicode(bool keepPercent)
     return wxString(L"\x2192");
 #endif
  /*
-  else if (m_textStyle == TS_SPECIAL_CONSTANT && m_text == wxT("d"))
+  else if (GetStyle() == TS_SPECIAL_CONSTANT && m_text == wxT("d"))
     return wxString(L"\x2202");
   */
 
@@ -695,6 +964,8 @@ wxString TextCell::GetSymbolUnicode(bool keepPercent)
       return wxString(L"e");
     else if (m_text == wxT("%i"))
       return wxString(L"i");
+    else if (m_text == wxT("%pi"))
+      return wxString(wxT("\x03C0"));
   }
 
   return wxEmptyString;
@@ -843,7 +1114,10 @@ wxString TextCell::GetSymbolSymbol(bool keepPercent)
     return "\xDB";
   else if (m_text == wxT(" xor "))
     return "\xC5";
+  else if (m_text == wxT("~>"))
+    return "\x219D";
 
+      
   if (!keepPercent) {
     if (m_text == wxT("%e"))
       return wxString(L"e");
